#!/usr/bin/python3
import time

#Make it easier to connect in thonny
print("Hello World")
time.sleep(5)
print("Loading...")

from json import dumps as json_encode
from GPIO import GPIO
from threadsafe import Message
import _thread
#import urequest # https://github.com/micropython/micropython-lib/blob/master/python-ecosys/urequests/urequests.py
import async_urequests as urequest
import uasyncio
from primitives import ESwitch, WaitAny
sleep=uasyncio.sleep
sleep_ms=uasyncio.sleep_ms

class debug:
	class net:
		errors=[]
		def log_error(err):
			print("Net Error:",err)
			errors.append([err,time.time()])
			if len(errors) > 3:
				errors.pop(0)
		def get_errors():
			log=[]
			now=time.time()
			for i in debug.net.errors:
				log.append([i[0],now-i[1]])
			return log

async def http(page,post=False,err=False):
	#print("http debug:",page,post,err)
	#return print("Simulation mode: ",page)
	# This function is ugly and i am not happy with it
	if not err:
		page="http://10.0.0.69:8080/"+page
		#print("URL =",page,"DATA =",post)
	if not GPIO.wlan.isconnected():
		debug.net.log_error("NOT CONNECTED")
		return False
	try:
		#print("http debug:",page,post)
		GPIO.light.onboard.on()
		if not post:
			r=await urequest.get(page)
		else:
			r=await urequest.post(page,data=json_encode(post),headers={"Content-type":"application/json"})
		GPIO.light.onboard.off()
		#print(page,':',r.status_code,"-",r.content)
	#except ValueError:
	#	#No idea why this happens, line is returned as []
	#	#My server log files disagree
	#	return True
	except Exception as e:
		GPIO.light.onboard.off()
		debug.net.log_error(e)
		#print(GPIO.nic.regs())
		return False
	#print(r.content)
	if r.status_code == 200 and await r.content == b"OK":
		await r.close()
		r=True
	else:
		debug.net.log_error([r.status_code,page])
		await r.close()
		r=False
	return r

async def wifi():
	from wifi_auth import ssid, password
	import network

	GPIO.wlan = network.WLAN(network.STA_IF)
	wlan=GPIO.wlan
	wlan.active(True)
	wlan.connect(ssid,password)
	wlan.config(pm = 0xa11140)# Power management is very very bad, ping time is around 1000x worse and packet loss insane

	while True:
		GPIO.light.onboard.on()
		while True:
			wstat=wlan.status()
			if wstat < 0 or wstat >= 3:
				break
			print('Waiting for WiFi connection...')
			GPIO.light.onboard.toggle()
			GPIO.light.offboard.toggle()
			await sleep(1)

		if wstat == 3:
			status = wlan.ifconfig()
			print('Wifi Connected; ip =',status[0])
			GPIO.light.onboard.off()
			GPIO.light.offboard.off()
			while GPIO.wlan.isconnected():
				await sleep(30)
		else:
			print("Failed to connect to wifi; retry in 30 seconds")
			await sleep(30)
		wlan.connect(ssid, password)

async def unblock(func, *args, **kwargs):#https://github.com/peterhinch/micropython-async/blob/master/v3/docs/THREADING.md#4-taming-blocking-functions
	def wrap(func, message, args, kwargs):
		message.set(func(*args, **kwargs))  # Run the blocking function.
	msg = Message()
	_thread.start_new_thread(wrap, (func, msg, args, kwargs))
	msg=await msg
	return msg

def readSensor(name):
	pin=GPIO.sensors.pins[name]["DHT"]
	try:
		pin.measure()
	except Exception as e:
		print(name,"DHT22 ERROR:",e)
		return False
	data={
		"temp":pin.temperature(),
		"water":pin.humidity()
	}
	#print(name, data)
	#if data["temp"] == 150.0 and data["water"] == 100 or data["temp"] == -50 and data["water"] == 0:
	# kitchen sensor likes to give strange bad data
	if data["temp"] > 100 or data["temp"] == -50 and data["water"] == 0:
		print(name,"DHT22 ERROR: I call BS;",data["temp"],"C;",data["water"],"%")
		return False# temp==150.0 and humidity == 100 if read too soon at power up
	else:
		return data

def readSensors(sensors):
	# Using core 2 on pico W does not crash so I may as well make use of it
	retry=[]
	i=-1
	while i < sensors["last_index"]:
		i+=1
		name=sensors["names"][i]
		if not GPIO.sensors.pins[name]["pwr"].value():
			GPIO.sensors.up(name)
			time.sleep(1)
		if GPIO.sensors.pins[name]["DHT"] is None:
			GPIO.sensors.pins[name]["DHT"]=GPIO.sensors.up(name)
			time.sleep(2)

		data=readSensor(name)
		if data:
			sensors["values"][ i ]=data
		else:
			GPIO.sensors.pins[name]["DHT"]=GPIO.sensors.down(name)
			retry.append({"n":name,"i":i})
	if (len(retry)>0):
		time.sleep(2)
	for r in retry:
		name=r["n"]
		GPIO.sensors.up(name)
		time.sleep(1)
		GPIO.sensors.pins[name]["DHT"]=GPIO.sensors.up(name)
		time.sleep(2)
		data=readSensor(name)
		if data:
			sensors["values"][ r["i"] ]=data
		else:
			sensors["values"][ r["i"] ] = False
			GPIO.sensors.pins[name]["DHT"]=GPIO.sensors.down(name)
	data={}
	for (name,val) in zip(sensors["names"], sensors["values"]):
		data[name]=val
	return data

def getReadings():
	return {
		#"bulb": GPIO.light.bulb.value(),
		"onboard": GPIO.light.onboard.value(),
		"offboard": GPIO.light.offboard.value(),
		"light": GPIO.light.value()
	}

async def buzzer():
	if GPIO.buzzer.lock:
		return
	GPIO.buzzer.lock=1
	while GPIO.buzzer.active:
		GPIO.buzzer.pin.duty_u16(32767)
		await sleep_ms(50)
		GPIO.buzzer.pin.duty_u16(0)
		await sleep_ms(50)
	GPIO.buzzer.lock=0

async def toggle_music(pin,i):
	url="music/outputs/outputs.php?set="+i+"&pico"
	btn=ESwitch(pin)
	evt=btn.open if i == "2" else btn.close
	while True:
		evt.clear()
		await evt.wait()
		uasyncio.create_task(http(url))

async def back_door(pin):
	sw=ESwitch(pin)
	def send_door(val):
		url="smart_home.php?update=living"
		json={"doors":{"back_door":int(val)},"status":getReadings()}
		uasyncio.create_task(http(url,json))
	send_door(pin.value())
	while True:
		sw.open.clear()
		sw.close.clear()
		send_door(await WaitAny((sw.open,sw.close)).wait() == sw.open)

async def webAPI(reader, writer):
	request_line = await reader.readline()
	print("\nClient connected:",request_line)
	if(request_line == b""):
		writer.write('HTTP/1.0 204 No Content')# is this right?
		await writer.drain()
		await writer.wait_closed()
		return print("Disconnect")

	length=0
	header=await reader.readline()
	while header != b"\r\n":
		if(header != b""):
			line=header.decode("utf-8")
			if line[0:15] == "Content-Length:":
				length=int(line[15:])
				print("Content length:",length)
			else:
				pass#print('Ignore header:',line)
		header=await reader.readline()

	request=request_line.decode("utf-8").split(' ')
	if request[0]=="GET":
		if request[1]=="/":
			json={}
			for i in GPIO.inputs:
				json[i]=GPIO.inputs[i].value()
			json=json_encode({
				"network":debug.net.get_errors(),
				"inputs":json,
				"outputs":getReadings()
			})
			writer.write('HTTP/1.0 200 OK\r\nContent-type: application/json\r\nContent-Length: '+str(len(json))+'\r\nCache-Control: no-cache\r\n\r\n'+json)
		else:
			writer.write('HTTP/1.0 404 Not Found\r\nContent-type: text/plain\r\nContent-Length: 3\r\n\r\n404')
	elif request[0]=="POST":
		data=await reader.readexactly(length)
		data=data.decode("utf-8")
		data=json_decode(data)
		print(data)
		if data["action"] == "buzzer":
			uasyncio.create_task(buzzer(data["value"]))
		#elif data["action"] == "blink":
		#	if not data["value"]:
		#		notifyControlActive=0
		#	elif not notifyControlActive:
		#		notifyControlActive=1
		#		uasyncio.create_task(notifyControl())
		writer.write('HTTP/1.0 200 OK\r\nContent-type: text/plain\r\nContent-Length: 2\r\nCache-Control: no-cache\r\n\r\nOK')
	else:
		writer.write("")
	try:
		await writer.drain()
		await writer.wait_closed()
		print("Client disconnected")
	except Exception as e:
		print("Error:",e)

async def main():
	while not GPIO.wlan.isconnected():
		await sleep(1)
	uasyncio.create_task(uasyncio.start_server(webAPI, "0.0.0.0", 80))
	uasyncio.create_task(toggle_music(GPIO.inputs["kitchen_button"],"1"))
	uasyncio.create_task(toggle_music(GPIO.inputs["living_button"],"2"))
	uasyncio.create_task(back_door(GPIO.inputs["back_door"]))

	json={"DHT22":{},"status":{}}

	sensors={
		"names":[],
		"values":[],
		"last_index":-1,
	}
	for i in GPIO.sensors.pins:
		sensors["names"].append(i)
		sensors["values"].append({})
		sensors["last_index"]+=1
	sensor_error=0
	await sleep(30)
	while True:
		data=await unblock(readSensors,sensors)
		uasyncio.create_task(http("smart_home.php?update=living",{
			"DHT22":data,
			"status":getReadings()
		}))
		await sleep(900)

#import os
# duplicate stdout and stderr to the log file - Catches errors and saves them.
#logfile = open('log.txt', 'a')
#os.dupterm(logfile)

uasyncio.create_task(wifi())

uasyncio.run(main())

